#include "LedControl.h"
#include <IRremote.h>

int RECV_PIN = 7;
IRrecv irrecv(RECV_PIN);
int CLK_PIN = 13;
int DATA_IN_PIN = 11;
int CS_PIN = 10;

int ***mat;
int i,j, k;
int side = 8;
int t_side = 4;
int no_dev = 3;
int max_dev = 4;
LedControl lc=LedControl(DATA_IN_PIN,CLK_PIN,CS_PIN,max_dev);

int tetromino[7][16]=
{
  {0,0,1,0,0,0,1,0,0,0,1,0,0,0,1,0},
  {0,0,1,0,0,1,1,0,0,0,1,0,0,0,0,0},
  {0,0,0,0,0,1,1,0,0,1,1,0,0,0,0,0},
  {0,0,1,0,0,1,1,0,0,1,0,0,0,0,0,0},
  {0,1,0,0,0,1,1,0,0,0,1,0,0,0,0,0},
  {0,1,0,0,0,1,0,0,0,1,1,0,0,0,0,0},
  {0,0,1,0,0,0,1,0,0,1,1,0,0,0,0,0}
};
int randomRows[7][8]=
{
  {1,0,1,0,1,0,1,0},
  {1,1,1,0,1,0,1,1},
  {1,0,0,1,1,0,1,0},
  {1,0,1,1,1,0,0,1},
  {1,1,0,0,1,0,0,1},
  {0,1,0,1,0,1,1,0},
  {1,0,0,0,1,0,0,1}
};
int nCurrentPiece = 0;
int nCurrentRotation = 0;
int nCurrentX = 0;
int nCurrentY = 0;
bool bForceDown = true;
bool bRotateHold = true;
int nScore = 0;
bool gameOver = true;
int **lines;
bool pauseGame = false;
bool cleared = false;
int a = -1;
int b = -1;
int Rotate(int px, int py, int r)
{
  int pi = 0;
  switch (r % 4)
  {
  case 0: // 0 degrees        // 0  1  2  3
    pi = py * 4 + px;         // 4  5  6  7
    break;                    // 8  9 10 11
                              //12 13 14 15

  case 1: // 90 degrees       //12  8  4  0
    pi = 12 + py - (px * 4);  //13  9  5  1
    break;                    //14 10  6  2
                              //15 11  7  3

  case 2: // 180 degrees      //15 14 13 12
    pi = 15 - (py * 4) - px;  //11 10  9  8
    break;                    // 7  6  5  4
                              // 3  2  1  0

  case 3: // 270 degrees      // 3  7 11 15
    pi = 3 - py + (px * 4);   // 2  6 10 14
    break;                    // 1  5  9 13
  }                           // 0  4  8 12

  return pi;
}
bool DoesPieceFit(int nTetromino, int nRotation, int nPosX, int nPosY)
{
  if(nTetromino == 0 && (nRotation ==2 || nRotation == 0)){
    if(nPosX >= 21)
      return false;
  } else {
    if(nPosX >= 22)
      return false;
  }
  for (int px = 0; px < 4; px++)
    for (int py = 0; py < 4; py++)
    {
      int pi = Rotate(px, py, nRotation);
      if (tetromino[nTetromino][pi] != 0 && mat[2-(nPosX+py)/side][(nPosY + px)%side][(nPosX + py)%side] != 0)
        return false;
    }

  return true;
}
void clearMat(){
  for(k = 0; k < no_dev; k++){
    for(i = 0; i < side; i++){
      for(j = 0;j < side; j++){
        mat[k][i][j] = 0;
      }
    }
  }
}
int countFullLines(){
  int c = 0;
  for(k = 0; k < no_dev; k++){
    for(i = 0; i < side; i++){
        if(lines[k][i] == 8){
          c++;
        }
    }
  }
  return c;
}

void setup() {
  Serial.begin(9600);
  irrecv.enableIRIn();
  lines = (int**)malloc(no_dev * sizeof(int*));
  mat = (int***)malloc(no_dev * sizeof(int**));
  for(i = 0; i < no_dev; i++){
    mat[i] = (int**)malloc(side * sizeof(int*));
    lines[i] = (int*)malloc(side * sizeof(int));
    for(j = 0; j < side; j++){
      mat[i][j] = (int*)malloc(side * sizeof(int));
      lines[i][j] = 0;
    }
  }
  for(k = 0; k < no_dev; k++){
    for(i = 0;i < side; i++){
      for(j= 0; j< side; j++){
          mat[k][i][j] = 0;
      }
    }
  }
  for(int address=0;address<max_dev;address++) {
    lc.shutdown(address,false);
    lc.setIntensity(address,8);
    lc.clearDisplay(address);
  }
  randomSeed(millis());
}


void loop() {
  if(!gameOver){
  if (IrReceiver.decode()) {
    if(IrReceiver.decodedIRData.command == 90 && pauseGame == false){
      if(DoesPieceFit(nCurrentPiece, nCurrentRotation, nCurrentX , nCurrentY- 1)){
        
       nCurrentY -= 1;
       if(nCurrentY <= -2){
          nCurrentY = -2;
        }
      }
    } else if(IrReceiver.decodedIRData.command == 8 && pauseGame == false){
      if(DoesPieceFit(nCurrentPiece, nCurrentRotation, nCurrentX , nCurrentY+ 1)){
        
        nCurrentY += 1;
        if(nCurrentY >= 5){
          nCurrentY = 5;
        }
      }
    }else if(IrReceiver.decodedIRData.command == 82 && pauseGame == false){
      if(DoesPieceFit(nCurrentPiece, nCurrentRotation, nCurrentX + 1, nCurrentY )){
        nCurrentX += 1;  
      }
    }else if(IrReceiver.decodedIRData.command == 24 && pauseGame == false){
      if(bRotateHold && DoesPieceFit(nCurrentPiece, nCurrentRotation + 1, nCurrentX, nCurrentY )){
        nCurrentRotation += 1;  
        bRotateHold = false;
      } else {
        bRotateHold = true;
      }
    }else if(IrReceiver.decodedIRData.command == 25){
                for(k = 0; k < no_dev; k++){
          for(i = 0;i < side; i++){
            for(j = 0;j < side; j++){
              Serial.print(mat[k][j][i]);
              Serial.print(" ");
            }
            Serial.println();
          }
        }
    } else if(IrReceiver.decodedIRData.command == 22){
       pauseGame = !pauseGame;
    } else if(IrReceiver.decodedIRData.command == 13){
      gameOver = true;
      a = -1;
      b = -1;
      cleared = false;
    }
    IrReceiver.resume();
  } 
    if(!pauseGame){
    if (DoesPieceFit(nCurrentPiece, nCurrentRotation, nCurrentX + 1, nCurrentY)){
          nCurrentX++;
    }
    else {
        for(i = 0;i < t_side; i++){
          for(j= 0; j< t_side; j++){
            if (tetromino[nCurrentPiece][Rotate(i, j, nCurrentRotation)] != 0){
              mat[2-(j+nCurrentX)/side][(i+nCurrentY)%side][(j+nCurrentX)%side] = 1;
            }
          }
        }
        nCurrentX = 0;
        nCurrentY = 0;
        nCurrentRotation = 0;
        nCurrentPiece = rand() % 7;
        for(k = 0; k < no_dev; k++){
          for(i = 0; i < side; i++){
            for(j = 0; j < side; j++){
              if(mat[k][j][i] == 1){
                lines[k][i]++;
              }
            }
          }
        }
        int pos_dev;
        int z,x,y,w;
        int g,h,i;
        int c = countFullLines();
        if(c != 0){
          nScore += (1 << c) * 100;
          for(k = 0; k < no_dev; k++){
            for(i = 0; i < side; i++){
              if(lines[k][i] == 8){
                for(j = 0; j < side; j++){
                  mat[k][j][i] = 0;
                }
                for(g = k; g < no_dev; g++){
                  for(h = i; h >= 0; h--){
                    for(j = 0; j < side; j++){
                      if(h == 0){
                        if(mat[g+1][j][side-1] != 0 && mat[g+1][j][side-1] != 1){
                          mat[g][j][h] = 0;
                        }else{
                          mat[g][j][h] = mat[g+1][j][side-1];
                        }
                      } else {
                        if(mat[g][j][h-1] != 0 && mat[g][j][h-1] != 1){
                          mat[g][j][h] = 0;
                        }else{
                          mat[g][j][h] = mat[g][j][h-1];
                        }
                      }
                    }
                  }
                }
                lines[k][i] = 0;
              }
            }
          }
        }
        
        delay(500);
        for(i = 0; i < no_dev; i++){
          for(j = 0; j < side; j++){
            lines[i][j] = 0;
          }
        }
        nScore += 25;
        gameOver = !DoesPieceFit(nCurrentPiece, nCurrentRotation, nCurrentX, nCurrentY);
        
        if(gameOver == false){
          Serial.println(nScore);
          a = -1;
          b = -1;
          cleared = false;
        }
    }
   
    }
  for(k = 0; k < no_dev; k++){
    for(i = 0; i < side; i++){
      for(j= 0; j< side; j++){
        if(mat[k][i][j] == 1){
          lc.setLed(k,i,j,true);
        } else {
          lc.setLed(k,i,j,false);
        }
      }
    }
  }
  
  for(i = 0;i < t_side; i++){
    for(j= 0; j< t_side; j++){
      if(tetromino[nCurrentPiece][Rotate(i,j,nCurrentRotation)] == 1){
         lc.setLed(2 - (j+nCurrentX)/side,(i+nCurrentY)%side,(j+nCurrentX)%side,true);       
      }
    }
  }
  
    delay(100);
  } else {
    if(!cleared){
      for(int address=0;address<max_dev;address++) {
        lc.clearDisplay(address);
      }
      cleared = true;
    }else {
      if(a != -1 && b != -1){
        if(b == 1){
          for(i = 0; i < side; i++){
            for(j = 0; j < side; j++){
              if(i == j || i == side-1 - j)
                lc.setLed(a,i,j,true);
            }
          }
        }
        if(b == 2){
          for(i = 0; i < side; i++){
            for(j = 0; j < side; j++){
              if(i != j && i != side-1 - j)
                lc.setLed(a,i,j,true);
            }
          }
        }
      }
      if (IrReceiver.decode()) {
        if(IrReceiver.decodedIRData.command == 69){
          if(a != -1){
            lc.clearDisplay(a);
          }
          a = 0;
          b = 1;
          clearMat();
        }
        if(IrReceiver.decodedIRData.command == 71){
          if(a != -1){
            lc.clearDisplay(a);
          }
          a =2;
          b = 1;
           clearMat();
        }
        if(IrReceiver.decodedIRData.command == 68){
          if(a != -1){
            lc.clearDisplay(a);
          }
          a = 3;
          b = 1;
          clearMat();
        }
        if(IrReceiver.decodedIRData.command == 64){
          if(a != -1){
            lc.clearDisplay(a);
          }
          a = 0;
          b = 2;
          clearMat();
        }
        if(IrReceiver.decodedIRData.command == 67){
          if(a != -1){
            lc.clearDisplay(a);
          }
          a = 1;
          b = 2;
          clearMat();
         
          for(i = 0; i < side; i++){
            int z = random(7);
            for(j = 0;j < side; j++){
              mat[0][j][i] = randomRows[z][j];
            }
          }
        }
        if(IrReceiver.decodedIRData.command == 7){
          if(a != -1){
            lc.clearDisplay(a);
          }
          a = 2;
          b = 2;
          clearMat();
        }
        if(IrReceiver.decodedIRData.command == 21){
          if(a != -1){
            lc.clearDisplay(a);
          }
          a = 3;
          b = 2;
          clearMat();
        }
        if(IrReceiver.decodedIRData.command == 9){
          if(a != -1){
            lc.clearDisplay(a);
          }
          a = -1;
          b = -1;
          clearMat();
        }
        if(IrReceiver.decodedIRData.command == 70){
          if(a != -1){
            lc.clearDisplay(a);
          }
          a = 1;
          b = 1;
          clearMat();
         
          for(i = 0; i < side; i++){
            int z = random(7);
            for(j = 0;j < side; j++){
              mat[0][j][i] = randomRows[z][j];
            }
          }
        }
        if(IrReceiver.decodedIRData.command == 28){
           for(i = 0; i < no_dev; i++){
            for(j = 0; j < side; j++){
              lines[i][j] = 0;
            }
          }
          for(i = 0; i < max_dev; i++){
            lc.clearDisplay(i);
          }
          nCurrentX = 0;
          nCurrentY = 0;
          nScore = 0;
          nCurrentRotation = 0;
          nCurrentPiece = random(8);
          gameOver = false;
        }
        IrReceiver.resume();
      }
    }
  }
}
